package cologne.eck.peafactory.peas;

/*
 * Peafactory - Production of Password Encryption Archives
 * Copyright (C) 2015  Axel von dem Bruch
 * 
 * This library is free software; you can redistribute it and/or modify it 
 * under the terms of the GNU General Public License as published 
 * by the Free Software Foundation; either version 2 of the License, 
 * or (at your option) any later version.
 * This library is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. 
 * See the GNU General Public License for more details.
 * See:  http://www.gnu.org/licenses/gpl-2.0.html
 * You should have received a copy of the GNU General Public License 
 * along with this library.
 */

/**
 * This is the base class for all Password Encrypted Archives (peas).
 * This is the parent class of all PswDialog* classes. 
 */


import java.awt.Window;
import java.io.*;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLClassLoader;
import java.nio.charset.Charset;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Locale;
import java.util.MissingResourceException;
import java.util.ResourceBundle;

import javax.swing.JFrame;
import javax.swing.JOptionPane;
import javax.swing.JPanel;

import settings.PeaSettings;
import cologne.eck.peafactory.crypto.CipherStuff;
import cologne.eck.peafactory.crypto.HashStuff;
import cologne.eck.peafactory.crypto.KeyDerivation;
import cologne.eck.peafactory.crypto.RandomStuff;
import cologne.eck.peafactory.peas.gui.PswDialogView;
import cologne.eck.peafactory.tools.Attachments;
import cologne.eck.peafactory.tools.Comparator;
import cologne.eck.peafactory.tools.Converter;
import cologne.eck.peafactory.tools.EntropyPool;
//import cologne.eck.peafactory.tools.Help;
import cologne.eck.peafactory.tools.ReadResources;
import cologne.eck.peafactory.tools.WriteResources;
import cologne.eck.peafactory.tools.Zeroizer;


public abstract class PswDialogBase { 

	// special panel for PswDialogView, currently only used in fileType:
	private static JPanel typePanel;
	
	private static String encryptedFileName;	
	
	// for file type and text file type: all files to select
	private static ArrayList<String> availableFileNames = new ArrayList<String>();
	
	private static String fileType; // set in daughter class
	
	private final static String PATH_FILE_NAME = PeaSettings.getJarFileName() + ".path";
	
	private static boolean insideJar = false;
	
	private static PswDialogBase dialog;
	
	private static char[] initializedPassword = null;
	
	private static String errorMessage = null;
	
	// i18n: 	
	private static String language; 
	private static Locale currentLocale;
	private static Locale defaultLocale = Locale.getDefault();
	private static ResourceBundle languagesBundle;
	
	/**
	 * The default character set, used whenever a string
	 * is converted to an array of bytes
	 */
	public static final Charset UTF_8 = Charset.forName("UTF-8");
	
	
	private static String workingMode = ""; // rescue mode or test mode
	
	//============================================================
	// --- abstract Methods:
	//

	// pre computation while windowOpening
	// this is nowhere used yet, but maybe used later 
	/**
	 * Used for pre-computations while the user types the password
	 */
	public abstract void preComputeInThread();
	//
	// before exit: clear secret values
	/**
	 * Overwriting secret value before closing the program
	 */
	public abstract void clearSecretValues();
	/**
	 * Start the decryption
	 */
	public abstract void startDecryption();	
	
	// used in menu of PswDialogView:
	/**
	 * Display a file chooser to initialize (encrypt)
	 * new files
	 */
	public abstract void initializeNewFiles();		
	
	// used in menu of PswDialogView:
	/**
	 * Display a file chooser to open
	 * previously encrypted files
	 */
	public abstract void openEncryptedFiles();	
	
	// currently only used in fileType:
	/**
	 * Get the chosen file names to decrypt
	 * 
	 * @return	an array of Strings of the selected file names
	 */
	public abstract String[] getSelectedFileNames();
	
	/**
	 * Get the version of the used PEA as String
	 * 
	 * @return the version as String
	 */
	public abstract String getVersion();
	
	/**
	 * Get the year of the publication of the used PEA
	 * 
	 * @return	year of publication as String
	 */
	public abstract String getYearOfPublication();
	
	/**
	 * Get the specific website of the used PEA, that is the 
	 * web address following the language, for example "/html/file_pea.html"
	 * 
	 * @return	the website as String
	 */
	public abstract String getWebsite();
	
	/**
	 * Get the link with the source code of the used PEA
	 * 
	 * @return	the link as String
	 */
	public abstract String getSourceLink();
	
	/**
	 * Get the frame with the decrypted content
	 * 
	 * @return	the frame of the opened PEA
	 */
	public abstract JFrame getLockFrame();

	//============================================================
	
	/**
	 * Set the language support for the PEA
	 * 
	 * @param _language					the language to set
	 * @throws MalformedURLException
	 */
	public final static void setLanguagesBundle(String _language) throws MalformedURLException {
		
		// i18n: 
		if (_language == null) {
			if (defaultLocale == null) { // openSuse
				defaultLocale = new Locale( System.getProperty("user.language") );
			}		
			language = defaultLocale.getLanguage(); // get default language
		} else {
			language = _language;
		}
	    currentLocale = new Locale(language); // set default language as language
	    
	    // If PswDialogBase is not used inside of a pea 
	    // FilePanel uses this bundle
	    try {
	    	
	    	languagesBundle = ResourceBundle.getBundle("resources.PeaLanguagesBundle", currentLocale);
	    	
	    } catch (MissingResourceException mre) {

	    	// If called in PeaFactory, not in PEA
	    	try {	    
	    		//System.out.println("Try to use bundle in folder resources (PeaFactory-Mode)");
	    	    File file = new File("resources");  
	    	    URL[] urls = {file.toURI().toURL()};  
	    	    ClassLoader loader = new URLClassLoader(urls);  
	    	    languagesBundle = ResourceBundle.getBundle("PeaLanguagesBundle", currentLocale, loader);
	    	    
	    	} catch  (MissingResourceException mre1) {

	    		System.err.println("Resource Bundle PeaLanguagesBundle "+ currentLocale + " not found");
	    		
	    		try{
	    			// for PeaFactory: en
	    			File file = new File("resources");  
		    	    URL[] urls = {file.toURI().toURL()};  
		    	    ClassLoader loader = new URLClassLoader(urls);  
		    	    currentLocale = new Locale("en"); 
		    	    language = "en";
		    	    languagesBundle = ResourceBundle.getBundle("PeaLanguagesBundle", currentLocale, loader);
	    		} catch (MissingResourceException mre2) {
	    			// for PEA: en
	    	    	languagesBundle = ResourceBundle.getBundle("resources.PeaLanguagesBundle", new Locale("en"));
	    		}
	    	}
	    }
	}
	
	/**
	 * Start the execution of the PEA: stop the collection of entropy, 
	 * add file to path file, start decryption process
	 * 
	 * @param rememberPath		remember the selected file names or not
	 */
	public final void startExecution(boolean rememberPath) {
		
		PswDialogView.setStarted(true);

		EntropyPool.getInstance().stopCollection();

		if (PswDialogView.isInitializing() == false) {

			// remember this file?
			if (rememberPath == true ) {
				
				String[] newNames;
				if (PswDialogBase.getFileType().equals("file")) {
					// set encrypted file names:
					newNames = PswDialogBase.getDialog().getSelectedFileNames();

				} else if (PswDialogBase.getFileType().equals("text file")
						|| PswDialogBase.getFileType().equals("image")) {
					int size = PswDialogBase.getAvailableFileNames().size();
					ArrayList<String> list = PswDialogBase.getAvailableFileNames();
					newNames = new String [size];
					for (int i = 0; i < size; i++) {
						newNames[i] = list.get(i);
					}					
					
				} else {
					newNames = new String[1];
					newNames[0] = PswDialogBase.getEncryptedFileName();
				}

				PswDialogBase.addFilesToPathFile( newNames );
			}
			if (PswDialogBase.getEncryptedFileName() == null 
					&& (PswDialogBase.getFileType().equals("text") )) { // only Notes 

				return;
			} else {
				PswDialogBase.getDialog().startDecryption();
			}

		} else { // internal File in directory "resources"


			PswDialogBase.getDialog().startDecryption();
	
			// remember this file?
			if (insideJar == false) {

				if (rememberPath == true ) {
					
					String[] newName = { PswDialogBase.getEncryptedFileName() };
					PswDialogBase.addFilesToPathFile( newName);
				}
			}
		} 
		PswDialogView.setStarted(false);
	}
	
	/**
	 * Derive the key from the password. Performs an initial
	 * hash and derives the key using the selected key derivation function
	 * 
	 * @param pswInputChars		the password
	 * @return					the derived key in required size
	 */
	public final static byte[] deriveKeyFromPsw(char[] pswInputChars) { 

		if (pswInputChars == null) {
			pswInputChars = "no password".toCharArray();
		} else if (pswInputChars.length == 0) {
			pswInputChars = "no password".toCharArray();
		}
		//System.out.println("PswDialogBase deriveKeyFromPsw\n"
		//		+ "  pswInputChars: " + new String(pswInputChars));

		byte[] pswInputBytes = Converter.chars2bytes(pswInputChars); 
		//Help.printBytes("PswDialogBase deriveKeyFromPsw \n  pswInputBytes", pswInputBytes);
		//Help.printBytes("PswDialogBase deriveKeyFromPsw \n  programRandomBytes", Attachments.getProgramRandomBytes());
		if (pswInputBytes.length > 0)  {
			Zeroizer.zero(pswInputChars);
		}
		// prevent using a zero password:
		pswInputChars = null;

		//=============
		//  derive key: 
		//	
		// 1. initial hash, reasons: 
		// - Bcrypt limited password length, 
		// - password remains n RAM in Bcrypt and Scrypt
		// - key file property
		Comparator.checkNullVector(pswInputBytes);
		byte[] pswHash = HashStuff.generateHKDFBytes(pswInputBytes, 
				Attachments.getProgramRandomBytes(), // if no key file property: this is a fixed string
				64); // size of pswHash: 64 bytes
		Zeroizer.zero(pswInputBytes);
		//Help.printBytes("PswDialogBase deriveKeyFromPsw \n  pswHash", pswHash);	

		Comparator.checkNullVector(pswHash);

		// 2. derive key from selected KDF: 
		byte[] keyMaterial = KeyDerivation.getKdf().deriveKey(pswHash);	
		
		Zeroizer.zero(pswHash);		
		return keyMaterial;
	}


	protected final byte[] getKeyMaterial() {

		//-------------------------------------------------------------------------------
		// Check if t is alive, if so: wait...
		if (PswDialogView.getThread() != null) {
			if (PswDialogView.getThread().isAlive()) { // this should never happen... 

				System.err.println("PswDialogBase: WhileTypingThread still alive!");
				try {
					PswDialogView.getThread().join();
				} catch (InterruptedException e) {
					if (PswDialogBase.getWorkingMode().equals("-t")){
						System.err.println("PswDialogBase: Joining thread interrupted");
						e.printStackTrace();
					}
				}
			}		
		}
	
		//---------------------------------------------------------------------
		// get password 
		char[] pswInputChars = null;
		if (PswDialogView.isInitializing() == false){
			pswInputChars = PswDialogView.getPassword();	
		} else {
			
			if (fileType.equals("text file") || fileType.equals("image")){
				pswInputChars = Arrays.copyOf(initializedPassword, initializedPassword.length);
			} else {
				pswInputChars = initializedPassword;
			}
			
		}
		PswDialogView.clearPassword();
//System.out.println("PswDialogBase getKeyMaterial, pswInputChars: " + new String(pswInputChars) );

		byte[] keyMaterial = deriveKeyFromPsw(pswInputChars);	
		
		if (fileType.equals("image") || fileType.equals("text file")){
			// do not zeroize
			// the password is needed to decrypt the image or 
			// to decrypt the file (text file)
		} else {
			Zeroizer.zero(initializedPassword);
			initializedPassword = null;
		}
//System.out.println("initialized Password: " + initializedPassword);
	
		if (keyMaterial == null) {
			PswDialogView.getView().displayErrorMessages("Key derivation failed.\n"
					+ "Program bug.", true);
			PswDialogView.clearPassword();
		}		
		return keyMaterial;
	}

	/**
	 * Initialize required values for execution: hash, cipher, 
	 * KDF, KDF parameters, programRandomBytes, salt, file identifier
	 */
	public final static void initializeVariables() {	
		
		HashStuff.setHashAlgo(PeaSettings.getHashAlgo() );
		HashStuff.setHMacAlgo(PeaSettings.getHashAlgo() );
		CipherStuff.setCipherAlgo(PeaSettings.getCipherAlgo() );
		KeyDerivation.setKdf( PeaSettings.getKdfScheme() );
		KeyDerivation.settCost(PeaSettings.getIterations() );
		KeyDerivation.setmCost(PeaSettings.getMemory() );
		KeyDerivation.setArg3(PeaSettings.getParallelization());
		KeyDerivation.setVersionString(PeaSettings.getVersionString() );
		Attachments.setProgramRandomBytes(PeaSettings.getProgramRandomBytes() );		
		Attachments.setFileIdentifier(PeaSettings.getFileIdentifier() );
	}
	
	// If a blank PEA is started first: 
	public final byte[] initializeCiphertext(byte[] ciphertext){
		// salt must be set manually and will be attached to the file later
		// in encryption step
		byte[] attachedSalt = new RandomStuff().createRandomBytes(KeyDerivation.getSaltSize());

		if ( ! fileType.equals("file") ){//PeaSettings.getBound() == false) {
			KeyDerivation.setSalt(attachedSalt);
		}
		
		// attach salt:
		ciphertext = Attachments.attachBytes(ciphertext, attachedSalt);
		// attach fileIdentifier
		ciphertext = Attachments.attachBytes(ciphertext, Attachments.getFileIdentifier());
		
		return ciphertext;
	}
	
	// if the PAE is not bonded to a content: the salt is attached
	public final void handleAttachedSalt(byte[] ciphertext) {

		byte[] tmp = new byte[KeyDerivation.getSaltSize()];
		System.arraycopy(ciphertext, 
				ciphertext.length 
				- Attachments.getFileIdentifierSize()
				- KeyDerivation.getSaltSize(), 
				tmp, 0, tmp.length);

		KeyDerivation.setSalt(tmp);
	}

	//== file functions: ===============================================================	

	/**
	 * Check a password for a cipher text: check if the pswIdentifier
	 * is the encrypted String "_PEA_FACTORY_ID_". 
	 * The nonce is taken from the cipher text, 
	 * the cipher text must in the original state,
	 * (not modified: (fileIdentifier, salt and nonce not cut)
	 * 
	 * @param cipherText	the original cipher text to check
	 * @param key			the key to use for decryption
	 * 
	 * @return				null if the password does not fail, 
	 * 						an error message otherwise
	 */
	public final static String checkPassword(byte[] cipherText, byte[] key) {
		
		// check the size of the cipher text:
		if (cipherText.length < (Attachments.getFileIdentifierSize() 
					+ KeyDerivation.getSaltSize()
					+ Attachments.getNonceSize() 
					+ Attachments.getPswIdentifierSize() ) ){
			return PswDialogBase.getBundle().getString("unsuitable_ciphertext_len");
		}
		
		// get pswIdentifier from cipher text:
		byte[] pswIdentifier = new byte[Attachments.getPswIdentifierSize()];
		System.arraycopy(cipherText,  cipherText.length 
					- Attachments.getFileIdentifierSize() 
					- KeyDerivation.getSaltSize()
					- Attachments.getNonceSize() 
					- Attachments.getPswIdentifierSize(),
				pswIdentifier,  0,  pswIdentifier.length);
		// get nonce from cipher text:
		byte[] nonce = new byte[ Attachments.getNonceSize()];
		System.arraycopy(cipherText,  cipherText.length 
					- Attachments.getFileIdentifierSize() 					
					- KeyDerivation.getSaltSize()
					- Attachments.getNonceSize(),  
				nonce,  0,  nonce.length);
		
		boolean result = Attachments.checkPswIdentifier(pswIdentifier, key, nonce);
		
		if (result == true) {
			return null;
		} else {
			return PswDialogBase.getBundle().getString("password_failed");
		}
	}
	
	/**
	 * Checks access to a file and the file identifier
	 * 
	 * @param fileName			the file to check
	 * @param addFileIfValid	add file to the list of available files if it is valid or not,
	 * 							this list is for EditorPEA, ImagePEA, NotesPEA - FilePea manages
	 * 							its own list of files in FileComposer
	 * @return					null if the file is valid, otherwise an error message
	 */
	public final static String checkFile(String fileName, boolean addFileIfValid) {
//System.out.println("PswDialogBase check: " + fileName);
		fileName = fileName.trim(); // ignore whitespace from manually editing the path file		
		File file = new File( fileName );		
	
		if (file.exists() && file.isFile() && file.canRead() && file.canWrite() ) {

			RandomAccessFile f = null;
			try {
				f = new RandomAccessFile(file, "rwd");
				//System.out.println("PswDialogBase checkFile: " + fileName);
		    	if ( Attachments.checkFileIdentifier(f, false) == true) {
			    	f.close();
			    	//============================
			    	// - - - Success: - - - - - - 
			    	//============================
			    	if (addFileIfValid == true) {
//System.out.println("PswDialogBase check file added: " + fileName);
			    		PswDialogBase.addAvailableFileName(fileName);
			    	}
			    	return null;
			    	
				} else {					
					f.close();
					//System.out.println("file identifier failed : " + fileName );
					return getBundle().getString("not_encrypted_with_this_archive") 
							+ " " + fileName;
				}
			} catch (FileNotFoundException e) {
				System.err.println(e.toString());
				return fileName + ": " + e;
			} catch (IOException e) {
				System.err.println(e.toString());
				return fileName + ": " + e;
			} catch (Exception e) {
				System.err.println(e.toString());
				try {
					f.close();
				} catch (IOException e1) {
					System.err.println(e1.toString());
				}
				return getBundle().getString("unexpected_error") + fileName + ": " + e;
			}
		} else {
			return getBundle().getString("no_access") + ": " + fileName;
		}
	}
	
	
	/**
	 * Used for PEAs which can open only one file at one time (simple notebook, image)
	 * and for editor (create CheckBoxes for valid files). 
	 * Searches the previously encrypted files. Checks the default location, 
	 * the external file name (set in PeaFactory) and the path file. 
	 * 
	 * @return	String "insideJar" if file is inside the PEA or null 
	 */
	public final static String searchSingleFile() {
		
		boolean addFileIfValid = true;
/*		if (fileType.equals("text file")){
			// do not add the file instead select by check boxes
			addFileIfValid = false;
		}*/
		
		PswDialogBase.setErrorMessage("");
				
		if (fileType.equals("image")){
			URL url = ReadResources.class.getClassLoader().getResource("resources/text.lock");
			try {
				//InputStream ios = ReadResources.class.getClass().getClassLoader().getResourceAsStream("resources/text.lock");
				InputStream ios = url.openStream();
				ios.close();
				PswDialogBase.addAvailableFileName(languagesBundle.getString("image_inside_jar_archive"));
			} catch (NullPointerException npe) {
				//System.err.println("No image inside the jar archive: " + npe.toString());			
			} catch (FileNotFoundException fnfe) {				
				//System.out.println("No image inside the jar archive: " + fnfe.toString());
			} catch (IOException ioe) {
				System.err.println("Image inside the PEA: access failed: " + ioe.toString());
			} catch (Exception e) {
				System.err.println("Image inside the PEA: access failed: " + e.toString());			
			}			
		}
		
		// 0. check default location
		// is automatically added in checkFile if valid
		checkFile("resources" + File.separator + "text.lock", addFileIfValid);

		// 1. check externalFilePath:
		if ( PeaSettings.getExternalFilePath() != null) {
			
			String externalFileName = PeaSettings.getExternalFilePath();

			if ( checkFile(externalFileName, addFileIfValid) == null ) { // access and fileIdentifier
				// everything handled in checkFile
			} else {
				PswDialogView.setMessage("###" + externalFileName + " - " + languagesBundle.getString("invalid_file"), true);
			}

		} 

		// 2. check path file
		String[] pathNames = accessPathFile();
//if (pathNames != null) for(int i=0;i<pathNames.length;i++) System.out.println("path file: " + pathNames[i]);
		
		if (pathNames == null) {
			//System.out.println("No result from previously saved file names in path file.");
		} else {

			for (int i = 0; i < pathNames.length; i++) {

				if (checkFile(pathNames[i], addFileIfValid) == null ) {
					// everything handled in checkFile
				} else {
					if (PswDialogBase.getWorkingMode().equals("-t")){
						System.err.println("invalid file in path file: " + pathNames[i]);
					}
				}
			}						
		}

		// check if at least one valid file was found:
		if (availableFileNames.size() == 0) {
			// no file found from standard location, external file name and path file:
			System.err.println("No valid file found.");
			PswDialogView.setMessage("###" + languagesBundle.getString("no_valid_file_found") 
					+ " -> " + languagesBundle.getString("open_or_initialize"), true);
		} 
		return null;
	}
	
	/**
	 * Checks the path file (file PEA's directory, name of PEA, with extension"path")
	 * for previously opened file names. 
	 * 
	 * @return	the names of the previously opened files (as an array of String)
	 * 			or null if the file does not exist, is empty or not accessible
	 */
	protected final static String[] accessPathFile() {
		File file = new File(PATH_FILE_NAME);
		if (! file.exists() ) {
			//System.err.println("PswDialogBase: no path file specified");
			return null;
		}
		if (! file.canRead() ) {
			System.err.println("PswDialogBase: can not read path file " + file.getName() );
			Window win = null;
			if (PswDialogView.getView().isVisible()){
				win = PswDialogView.getView();
			} else {
				win = PswDialogBase.getDialog().getLockFrame();
			}
			JOptionPane.showInternalMessageDialog(win,
					languagesBundle.getString("path_file_access_error")
				//"There is are file containing names of potentially encrypted files,\n" +
				//" but no access to it: \n"
				+ PATH_FILE_NAME, 
				"Info",//title
				JOptionPane.INFORMATION_MESSAGE);
			return null;
		}
		
		byte[] pathBytes = ReadResources.readExternFile( file.getName() );
		if (pathBytes == null) {
			System.err.println("PswDialogBase: path file does not contain any file name: " + file.getPath() );
			return null;
		}
		String pathString = new String(pathBytes, UTF_8);

		String[] pathNames = pathString.split("\n");
	
		return pathNames;			
	}
	

	/**
	 * Add selected files to the path file
	 * (Checks if files already exists in path file).
	 * 
	 * @param selectedFileNames		the file names to add to the path file
	 * 								as an array of Strings
	 */
	public final static void addFilesToPathFile(String[] selectedFileNames) {

		byte[] newPathBytes = null;
		StringBuilder newPathNames = null;
				
		File file = new File(PATH_FILE_NAME);
		
		if (! file.exists() ) { // WriteResources creates new file
			
			newPathNames = new StringBuilder();
			
			for (int i = 0; i < selectedFileNames.length; i++) {
				newPathNames.append(selectedFileNames[i]);
				newPathNames.append("\n");
			}
			
		} else {	// path file exists
		
			if (! file.canRead() || ! file.canWrite() ) {
				System.err.println("PswDialogBase: can not read and write path file " + file.getName() );
				Window win = null;
				if (PswDialogView.getView().isVisible()){
					win = PswDialogView.getView();
				} else {
					win = PswDialogBase.getDialog().getLockFrame();
				}
				JOptionPane.showInternalMessageDialog(win,
						languagesBundle.getString("path_file_access_error")
					+ PATH_FILE_NAME, 
					"Info",//title
					JOptionPane.INFORMATION_MESSAGE); 
				return;
			}
			
			// get file names from path file:
			String[] oldPathNames = accessPathFile();
			newPathNames = new StringBuilder();
			
			if ( file.length() == 0) { // empty file
				
				System.out.println("empty path file");
				
				//append new file names: 
				for (int i = 0; i < selectedFileNames.length; i++) {
						newPathNames.append(selectedFileNames[i]);
						newPathNames.append("\n");
				}
			
			} else {				
				
				// append old file names:
				for (int i = 0; i < oldPathNames.length; i++) {
					newPathNames.append(oldPathNames[i]);
					newPathNames.append("\n");
				}
				// check if file name already exists, if not append
				for (int i = 0; i < selectedFileNames.length; i++) {
					boolean append = true;
					for (int j = 0; j < oldPathNames.length; j++) {
						if (selectedFileNames[i].equals(oldPathNames[j] )){
							append = false;
						}
					}
					if (append == true) {
						newPathNames.append(selectedFileNames[i]);
						newPathNames.append("\n");
					}
				}
			}
		}

		newPathBytes = new String(newPathNames).getBytes(UTF_8);
			
		WriteResources.write(newPathBytes, PATH_FILE_NAME, null);		
	}
	

	/**
	 * Checks if current encryptedFileName is in path file.
	 * If not: asks to remember the file and store in path file.
	 * 
	 * @param newFileName	the file name to check
	 * @return				true if the path file was updated with 
	 * 						new file name
	 */
	public final static boolean pathFileCheck(String newFileName) {

		if (insideJar == true) {
			return false;
		}
		
		if (newFileName != null) {
			// path stored inside the PEA
			if (newFileName.equals(PeaSettings.getExternalFilePath() )) {
				return false; // no need to remember
			}			
			// default location
			if (newFileName.equals("resources" + File.separator + "text.lock")){
				return false; // no need to remember			
			}
		}

		String pathFileName = PeaSettings.getJarFileName() + ".path";
		byte[] pathFileContent = ReadResources.readExternFile(pathFileName);
		if (pathFileContent == null) {
			int n = JOptionPane.showConfirmDialog(getDialog().getLockFrame(),
					languagesBundle.getString("remember_files")
					+ languagesBundle.getString("path_file_storage_info")
			+ pathFileName, 
				"?", 
				JOptionPane.YES_NO_OPTION, 
				JOptionPane.PLAIN_MESSAGE);
				if (n == 2) { // cancel
					return false;
				} else if (n == 0){
					
					if (PswDialogBase.getFileType().equals("text file")
							|| PswDialogBase.getFileType().equals("image")){
						// more than one file
						int len = availableFileNames.size();
						String allFileNames = "";
						for (int i = 0; i < len; i++) {
							
							allFileNames += availableFileNames.get(i) +"\n";
						}
						byte[] encryptedFileNameBytes = allFileNames.getBytes(UTF_8);

						WriteResources.write(encryptedFileNameBytes, pathFileName, null);

						
					} else {
					
						byte[] encryptedFileNameBytes = newFileName.getBytes(UTF_8);

						WriteResources.write(encryptedFileNameBytes, pathFileName, null);
					}
				}
			return true;
		} else {
			String pathNamesString = new String(pathFileContent, UTF_8);

			String[] pathNames = pathNamesString.split("\n");
			// check if encryptedFileName is already included
			for (int i = 0; i < pathNames.length; i++) {
				if (pathNames[i].equals(newFileName)) {
					//System.out.println("Already included: " + encryptedFileName);
					return false;
				}
			}
						
			// is not included:
			Window win = null;
			if (PswDialogView.getView().isVisible()){
				win = PswDialogView.getView();
			} else {
				win = PswDialogBase.getDialog().getLockFrame();
			}
			int n = JOptionPane.showConfirmDialog(win,
			languagesBundle.getString("remember_files")
			+ languagesBundle.getString("path_file_storage_info")
			+ pathFileName , 
			"?", 
			JOptionPane.YES_NO_OPTION, 
			JOptionPane.PLAIN_MESSAGE);
			if (n == 2) { // cancel
				return false;
			} else if (n == 0){ // ok
				
				//store in path file:
				byte[] encryptedFileNameBytes =  ("\n" + newFileName).getBytes(UTF_8);

				byte[] newContent = new byte[pathFileContent.length + encryptedFileNameBytes.length];
				System.arraycopy(pathFileContent, 0, newContent, 0, pathFileContent.length);
				System.arraycopy(encryptedFileNameBytes, 0, newContent, pathFileContent.length, encryptedFileNameBytes.length);
				
				if (newContent != null) {
					WriteResources.write(newContent, pathFileName, null);
				}
				return true;
				
			} else {
				return false;
			}
		}		
	} 

	/**
	 * Checks if file names are already stored in the path file.
	 * If not: asks to remember the files and store.
	 * 
	 * @param newFileNames	the file names to check
	 * @return				the file names which were added to
	 * 						the path file or null if no file name was added
	 */
	public final static ArrayList<String> pathFileCheck(ArrayList <String> newFileNames) {
		
		if (newFileNames == null) {
			return null;
		}
		// return value and list of files to add to path file
		ArrayList<String> updatedFileNames = null;

		// the name of the path file where file names are stored to remember:
		String pathFileName = PeaSettings.getJarFileName() + ".path";
		// the current content of the path file:
		byte[] pathFileContent = ReadResources.readExternFile(pathFileName);
		if (pathFileContent == null) {
			pathFileContent = " ".getBytes(getCharset());
		}
		// the current content as String
		String pathNamesString = new String(pathFileContent, getCharset());
		// file names are separated by \n: String array of current file names
		// in path file
		String[] pathNames = pathNamesString.split("\n");
		int len = newFileNames.size();
		
		// check each new file name:
		for (int i = 0; i < len; i++) {					
			// path stored inside the PEA
			if (newFileNames.get(i).equals(PeaSettings.getExternalFilePath() )) {
				// no need to remember
			} else if (newFileNames.get(i).equals("resources" + File.separator + "text.lock")){
				// no need to remember			
			} else {		
				// check against every file name of path file
				boolean inPathFile = false;
				for (int j = 0; j < pathNames.length; j++) {
					if (pathNames[j].equals(newFileNames.get(i))) {
						//System.out.println("Already included: " + newFileNames.get(i));
						inPathFile = true;
						break;
					} else {
						//
					}
				}
				if (inPathFile == false){ // no match found
					// file name is not in path file:
					if (updatedFileNames == null){
						updatedFileNames = new ArrayList<String>();
					}
					// add to new list:
					updatedFileNames.add(newFileNames.get(i));
				}
			}
		}
		if (updatedFileNames != null) { // at least one file was not in path file

			Window win = null; // the owner of the JOptionPane
			if (PswDialogView.getView().isVisible()){
				win = PswDialogView.getView();
			} else {
				win = PswDialogBase.getDialog().getLockFrame();
			}
			int n = JOptionPane.showConfirmDialog(win,
			languagesBundle.getString("remember_files")
			+ languagesBundle.getString("path_file_storage_info")
			+ pathFileName , 
			"?", 
			JOptionPane.YES_NO_OPTION, 
			JOptionPane.PLAIN_MESSAGE);
			if (n == 0){ // ok				
				//store in path file:
				int updatedLen = updatedFileNames.size();
				//String newContent = "";
				for (int i = 0; i < updatedLen;i++) {
					// separate by \n and add to current list
					pathNamesString += "\n" + updatedFileNames.get(i);
				}
				
				if (pathNamesString.trim().length() > 1) { // 
					byte[] newContent = pathNamesString.getBytes(getCharset());
					WriteResources.write(newContent, pathFileName, null);
				}
				
			} else { // cancel,close
				return null;
			}
		}
		return updatedFileNames;
	} 

	
	//==========================================
	// Getter & Setter

	public final static Charset getCharset(){
		return UTF_8;
	}
	
	protected final static void setFileType(String _fileType) {
		fileType = _fileType;
	}
	public final static String getFileType() {
		return fileType;
	}
	public final static String getEncryptedFileName() {
		return encryptedFileName;
	}
	public final static void setEncryptedFileName(String _encryptedFileName) {
		encryptedFileName = _encryptedFileName;
	}
	protected final static void setDialog(PswDialogBase _dialog) {
		dialog = _dialog;
	}
	public final static PswDialogBase getDialog() {
		return dialog;
	}
	public final static String getPathFileName() {
		return PATH_FILE_NAME;
	}
	public final static JPanel getTypePanel() {
		return typePanel;
	}
	public final static void setTypePanel( JPanel panel){
		typePanel = panel;
	}
	public final static String getErrorMessage() {
		return errorMessage;
	}
	public final static void setErrorMessage(String newErrorMessage) {
		errorMessage = newErrorMessage;
	}
	public final static ResourceBundle getBundle(){
		if (languagesBundle == null) { // set default language
			try {
				setLanguagesBundle(null);
			} catch (MalformedURLException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		}
		return languagesBundle;
	}
	
	/**
	 * Set the password from initialization process
	 * to use in key derivation
	 * 
	 * @param psw	the password to set
	 */
	public void setInitializedPassword(char[] psw){
		initializedPassword = psw;
	}
	/**
	 * @return the workingMode
	 */
	public static String getWorkingMode() {
		return workingMode;
	}
	/**
	 * @param workingMode the workingMode to set
	 */
	public static void setWorkingMode(String workingMode) {
		PswDialogBase.workingMode = workingMode;
	}
	/**
	 * Return the file name which can be decrypted with his PEA
	 * 
	 * @return the availableFileNames
	 */
	public static ArrayList<String> getAvailableFileNames() {
		return availableFileNames;
	}
	/**
	 * Set the available file names
	 * 
	 * @param newAvailableFileNames the file names to set
	 */
	public static void setAvailableFileNames(ArrayList<String> newAvailableFileNames) {
		availableFileNames = newAvailableFileNames;
	}
	/**
	 * @param availableFileNames the availableFileNames to set
	 */
	public static void addAvailableFileName(String availableFileName) {
		
		if (fileType.equals("file")){
			return;
		}

		if (PswDialogBase.availableFileNames.contains(availableFileName)){
//			System.out.println("file already available: " + availableFileName);
		} else {
			PswDialogBase.availableFileNames.add(availableFileName);
			PswDialogView.addAvailableFile(availableFileName, false);// false: not new selected
			//Collections.sort(PswDialogBase.availableFileNames);
		}
	}
	/**
	 * Check if the content is stored as file inside the jar archive,
	 * used in LockFrameImage
	 * 
	 * @return 	true is the content is inside the jar file
	 */
	public static boolean isInsideJar() {
		return insideJar;
	}
	/**
	 * Set variable to check if the content is located
	 * inside the jar archive, 
	 * used in PswDialogImage
	 * 
	 * @param _inside
	 */
	public static void setInsideJar(boolean _inside){
		insideJar = _inside;
	}

}
